
/* Twiddler Driver for Windows NT (Test User Mode Version)
 * 1/15/97
 * written by:Jeff Levine
 * <tornado@media.mit.edu>
 * for Handykey Corporation 
*/

#include "twiddler.h"

// tilt-sensor, linearizer.. copied from Win3.1 version of driver
USHORT LIN_ANGLE[] = {
	   0,  1,  2,  3,  4, 5, 6, 7, 9, 11,      //  00
	  13, 15, 16, 17, 18, 19, 20, 21, 22, 23,      //	; 10
	  24, 25, 26, 27, 28, 29, 30, 31, 32, 33,      //	; 20
	  34, 35, 36, 37, 38, 39, 40, 41, 42, 43,      //	; 30
	  44, 45, 46, 47, 48, 49, 50, 51, 52, 53,      //	; 40
	  54, 55, 56, 57, 57, 58, 58, 59, 59, 60,      //	; 50
	  60, 61, 61, 61, 62, 62, 63, 63, 63, 63,      //	; 60
	  64, 64, 65, 65, 65, 65, 66, 66, 66, 66,      //	; 70
	  67, 67, 67, 67, 68, 68, 68, 69, 69, 69,      //	; 80
	  69, 69, 69, 70, 70, 70, 70, 71, 71, 71,      //	; 90
	  71, 71, 71, 72, 72, 72, 72, 72, 72, 72,      //	; 100
	  73, 73, 73, 73, 73, 74, 74, 74, 74, 74,      //	; 110
	  74, 74, 74, 74, 75, 75, 75, 75, 75, 75,      //	; 120
	  75, 76, 76, 76, 76, 76, 76, 76, 76, 76,      //	; 130
	  76, 76, 77, 77, 77, 77, 77, 77, 77, 77,      //	; 140
	  77, 77, 77, 77, 77, 77, 78, 78, 78, 78,      //	; 150
	  78, 78, 78, 78, 78, 78, 78, 78, 78, 78,      //	; 160
	  78, 78, 78, 79, 79, 79, 79, 79, 79, 80,      //	; 170
	  80, 80, 81, 81, 81, 81, 82, 82, 83, 83,      //	; 180
	  84, 84, 85, 85, 86, 86, 87, 87, 88, 88,      //	; 190
	  89, 89, 89, 89, 90, 90, 90, 90, 90, 90,      //	; 200
	  90, 90, 90, 90, 90, 90, 90, 90, 90, 90,      //	; 210
	  90, 90, 90, 90, 90, 90, 90, 90, 90, 90,      //	; 220
	  90, 90, 90, 90, 90, 90, 90, 90, 90, 90,      //	; 230
	  90, 90, 90, 90, 90, 90, 90, 90, 90, 90,      //	; 240
	  90, 90, 90, 90, 90, 90};      //	; 250

int lastvert = 0;
int lasthoriz = 0;
int ttyfd,fifo;
HANDLE twid_port;
int xterm_on, mouse_on, mouse_reverse;
float mouse_res_x,mouse_res_y;
struct keylist **table; /* everybody needs to know */
struct dxdy *mouse_filter;
int filter_index;
struct state states;
BYTE shift_scan,control_scan, alt_scan;
HANDLE  hServerStopEvent = NULL;

void main (int argc, char **argv)
{
  int i;
  int numread;
  unsigned char buffer[5];
  COMMTIMEOUTS timeout;

  printf ("Handykey Twiddler NT driver (test user mode) version 2.1\n\n");
  printf ("Note: This program has been stripped of all NT specific\n");
  printf ("code, and should run on Windows 95.  The driver will quit\n");
  printf ("when this window is closed. (Minimize for convenience)\n\n");

  states.flags = 0;
  if (!init (&states,argc,argv))
    {
      return;
    }
  
  norepeat.num=3;
  norepeat.code = calloc (3,sizeof(UINT));
  norepeat.code[0] = VK_CAPITAL;
  norepeat.code[1] = VK_NUMLOCK;
  norepeat.code[2] = VK_SCROLL;

  if (!GetCommTimeouts(twid_port,&timeout))
    {
      CloseHandle(twid_port);
      return;
    }
  timeout.ReadIntervalTimeout = 0;
  timeout.ReadTotalTimeoutMultiplier = 0;
  timeout.ReadTotalTimeoutConstant = 0;
  if (!SetCommTimeouts (twid_port,&timeout))
    {
      CloseHandle (twid_port);
      return;
    }
  
  
  while (1)
    {
//      uh_oh = WaitForSingleObject(hServerStopEvent,0);
//      if (uh_oh == WAIT_OBJECT_0)
//	return;  /*service is stopped */
      for (i=0;i<5;i++)
	{
	  ReadFile (twid_port, buffer+i, 1,&numread,NULL);
	  if (numread !=1)
	    {
	      i--;
	    }
	  else if ((i==0)&&(buffer[0]&0x80))
	    i--; /* we're out of sync... let's keep trying */
	  else if ((i>0) && (!buffer[i]&0x80))
	    {
	      buffer[0]= buffer[i];
	      i=1;
	    }    /* Found a header byte in the middle of a sequence! */
	  /* start over with that byte first */ 
	}
#ifdef DEBUG
      printf ("Buffer Contents: ");
      for (i=0;i<5;i++)
	printf ("%x ",buffer[i]);
      printf ("\n");
#endif 
      processbuffer (buffer);
    }
}

void processbuffer (unsigned char *buffer)
{
  
  /* event enabled upon key press
     event generated upon key release
     unless repeat (then it waits for a time and the generates an event */
  STATE events;
  int index=0;
  char mouse_but;
  int dx,dy,cx,cy,angy,angx,meancy,meancx,fil_lev;

  index |= ((buffer[0]&0x7f) | ((buffer[1]&0x3f)<<7));
  
  events = &states;
  events->last_state = events->cur_state;
  events->cur_state.index = index;
  
  if (events->cur_state.index>events->last_state.index)
    (events->flags |= EVENT_ENABLE);
  
  if (events->time_count<REPEAT_TIME_THRESH)
    events->time_count++;
  /* handle mouse togggle */
  mouse_but = buffer[1]&0x40;

  events->cur_state.horiz_tilt = 0 | (((buffer[3]&0x7c)>>2) | 
				      ((buffer[4]&0x07)<<5));
  events->cur_state.vert_tilt = 0 | ((buffer[2]&0x7f) | 
				     ((buffer[3]&0x01)<<7));
  if (!(buffer[3]&0x02))
    events->cur_state.vert_tilt += 256;
  if (!(buffer[4]&0x08))
    events->cur_state.horiz_tilt += 256;
  /* flip the high order bit to get offset binary,
     then just use integer subtract... etc.. */
    
  /* cx, cy is the RAW data. */
  cx = -(events->cur_state.horiz_tilt - 256);
  cy = -(events->cur_state.vert_tilt - 256);
  if (mouse_reverse) cy=-cy;
  
  // ok, now store raw number, get 3-ave filter, then linearize that #
//  printf ("%d %d %d\n",filter_index, cx, cy);

  get3mean (mouse_filter, &meancx,&meancy,cx,cy,&filter_index,
	    FILTER_LEVEL);      
  if (meancx<0)
    angx = -LIN_ANGLE [-meancx];
  else
    angx = LIN_ANGLE [meancx];
  if (meancy<0)
    angy = -LIN_ANGLE [-meancy];
  else
    angy = LIN_ANGLE [meancy];

  dx = angx - lasthoriz;
  dy = angy - lastvert;

  add_to_filter (mouse_filter, dx,dy,cx,cy,&filter_index,
		 FILTER_LEVEL);  
  lasthoriz = angx;
  lastvert = angy;

  if (abs(angx)<abs(angy))
    fil_lev =  abs(angy);
  else fil_lev = abs(angx);
  fil_lev = (fil_lev/20)+5;
  if (mouse_but)
    {
      handle_mouse_activity (index, buffer, fil_lev); 
      return;
    }

  if (!(events->flags&EVENT_ENABLE))
    return ; /* if key events not enabled, let's return*/
  
  
  /* if the last state was a mapped character and we just released */
  if (events->cur_state.index < events->last_state.index)
    {
      /* generate event based on the last state */
      if (events->repeat_count>=REPEAT_THRESH)/* we just repeated some stuff */
	events->flags &= ~REPEAT_ENABLE;
      else
	events->flags |= REPEAT_ENABLE; /* else leave repeat option open */
      events->time_count = 0;
      events->repeat_count=0; 
      events->last_event=events->last_state;
      events->flags &= ~EVENT_ENABLE;
      get_text();
      return;
    }
  
  /* we just pressed something, lets check if we're ready
     to go into repeat mode */

  if ((events->flags&REPEAT_ENABLE) && 
      (events->time_count-events->repeat_count<REPEAT_TIME_THRESH) &&
      (search_table (events->cur_state.index)) &&
      (events->cur_state.index==events->last_event.index))
    events->repeat_count++; /* if state is a repeat state, and not first instance */
  else events->repeat_count=0;
  if (events->repeat_count>REPEAT_THRESH)
    events->repeat_count=REPEAT_THRESH; 
  if (events->repeat_count == REPEAT_THRESH)
    {
      /* generate an event based on the current state */
      events->last_event=events->cur_state;
      get_text();
    }
  return;
}

void handle_mouse_activity (int index, unsigned char *buffer,int fil_lev)
  
{
  static short leftdown,middledown=FALSE,rightdown=FALSE;
  static int mouse_but_state = M_OFF;
  static int lastvert = 0;
  static int lasthoriz = 0;
//  int delta;
  int dx,dy;
  int mouse_res;
  STATE events;

  events = &states;
  
  switch (index&0x30)
    {
    case (MOUSE_HIRES):
      mouse_res=4;
      break;
    case (MOUSE_LORES):
      mouse_res=8;
      break;
    default:
      mouse_res=2;
    }
  if (!mouse_on)
    mouse_on=1;


  filter_mouse (mouse_filter,fil_lev,mouse_res,&dx,&dy,
		&filter_index,FILTER_LEVEL);      

  if ((dx || dy) || mouse_res ==8)
    mouse_event(MOUSEEVENTF_MOVE,dx,dy,0,0); 

  /* move mouse*/
  if (!(events->flags&EVENT_ENABLE))
    return ; /* if key events not enabled, let's return */
    
    
  /* check to get special mouse events */
  index &= 0xff;
  if (index==0)
    {
      events->flags &= ~EVENT_ENABLE;
      if (leftdown)
	mouse_event (MOUSEEVENTF_LEFTUP,0,0,0,0);
      if (middledown)
	mouse_event (MOUSEEVENTF_MIDDLEUP,0,0,0,0);
      if (rightdown)
	mouse_event (MOUSEEVENTF_RIGHTUP,0,0,0,0);
      leftdown=middledown=rightdown=FALSE;
      return ;
    }
  if (events->cur_state.index > events->last_state.index)
    switch (index&0xcf)
      {
      case (LEFT_BUTTON):
	mouse_event (MOUSEEVENTF_LEFTDOWN,0,0,0,0);
	leftdown = TRUE;
	break;
      case (MID_BUTTON):
	mouse_event (MOUSEEVENTF_MIDDLEDOWN,0,0,0,0);
        middledown=TRUE;
	break;
      case (RIGHT_BUTTON):
	mouse_event (MOUSEEVENTF_RIGHTDOWN,0,0,0,0);
	rightdown=TRUE;
	break;
      case (M_DOUBLE_CLICK):
	if (!leftdown)
	  mouse_event (MOUSEEVENTF_LEFTDOWN,0,0,0,0);
	mouse_event (MOUSEEVENTF_LEFTUP,0,0,0,0);
	mouse_event (MOUSEEVENTF_LEFTDOWN,0,0,0,0);
	mouse_event (MOUSEEVENTF_LEFTUP,0,0,0,0);
	leftdown=FALSE;
      }

  if (events->cur_state.index < events->last_state.index)
    {
      switch (events->last_state.index)
	{
	case (MOUSE_REVERSE):
	  events->flags &= ~EVENT_ENABLE;
	  mouse_reverse = 1-mouse_reverse;
	  return ;
	default:
	  break;
	  /* we dont have any other keys to be pressed during mouse mode,
	     but here's where normal style "hardwired" key presses 
	     like mouse_reverse would go */
	}
    }
  return;
} 

void get3mean (struct dxdy *filter, int *cx, int *cy, int newcx, int newcy,
	       int *index, int filter_size)
{
  int i;
  *cx = 0;
  *cy = 0;
  (*index)--;

  if (*index==-1)
    *index = (filter_size-1);
  for (i=0;i<2; i++)
    {
      *cx += filter[*index].dxraw;
      *cy += filter[*index].dyraw;
      if (i==0)
	(*index)++;

      if (*index==filter_size)
	*index=0;
    }
  *cx += newcx;
  *cy += newcy;
  *cx /= 3;
  *cy /= 3;
}

void add_to_filter (struct dxdy *filter, int dx, int dy, int cx,int cy,
		    int *index, int filter_size)

{
  (*index)++;
  if (*index==filter_size) *index=0;
  filter[*index].dx = dx;
  filter[*index].dy = dy;
  filter[*index].dxraw = cx;
  filter[*index].dyraw = cy;
}
void filter_mouse (struct dxdy *filter, int filter_level,
			   int res_level, int *dx, int * dy, int *index,
			   int filter_size)
{
  int i;
  
  /* add dynamic filtering and 
     dynamic resolution */
  *dx = 0;
  *dy = 0;
  for (i=0;i<(filter_level-1); i++)
    {
      (*index)--;
      if (*index==-1)
	*index = (filter_size-1);
    }
  for (i=0;i<filter_level; i++)
    {
      *dx += filter[*index].dx;
      *dy += filter[*index].dy;
      if (i<(filter_level-1))//dont increment past last data
	(*index)++;
      if (*index==filter_size)
	*index=0;
    }
  *dx*=7;
  *dy*=7;
  *dx /= (res_level*filter_level);
  *dy /= (res_level*filter_level);
}

void get_text ()
{
  STATE events;
 
  int MOD_FLAGS=0;
  char line2[MAXLINE]="",line[MAXLINE]="";
  struct keylist *temp;
  int i;
  events = &states;

  temp = search_table (events->last_event.index);
  if (temp  && (temp->vkeys[0]==VK_NUMLOCK) && (!(temp->vkeys[0]&0x100)))
    { /* NUMLOCK PRESSED */
      if (states.flags & NUM_LOCK)
	{
	  states.flags &= (~NUM_LOCK);
	}
      else
	{
	  states.flags |=NUM_LOCK;
	}
    }
  else
    {  if (states.flags &NUM_LOCK)
      {
	events->last_event.index |= NUM_INDEX;
	temp = search_table(events->last_event.index);
      }
    }
  if (temp==NULL)
    {
      if (events->last_event.index &CTRL_INDEX)
	{
	  MOD_FLAGS |= 0x02;
	}
      if (events->last_event.index &(ALT_INDEX))
	{
	  /* add alt down */
	  MOD_FLAGS |= 0x04;
	}
      if (events->last_event.index &(SHIFT_INDEX))
	{
	  /* add shift down */
	  MOD_FLAGS |= 0x01;
	}
      if (MOD_FLAGS)
	{
	  if (states.flags &NUM_LOCK)
	    temp =search_table((events->last_event.index&0xff)|NUM_INDEX);
	  else
            temp =search_table(events->last_event.index&0xff);
	  if (temp)
	    {
	      for (i=0;i<norepeat.num;i++)
		if (temp->vkeys[0] ==(norepeat.code[i]))
		  {      
		    events->flags &= ~REPEAT_ENABLE;
		  }
	      generate_key_events (temp, MOD_FLAGS);
	    }
	}
    }
  else
    {
      if ((temp->vkeys[0] == VK_NUMLOCK) && (!(temp->vkeys[0] &0x100)))
	{   
	
	}
      for (i=0;i<norepeat.num;i++)
	if (temp->vkeys[0] ==(norepeat.code[i]))
	  {
	    events->flags &= ~REPEAT_ENABLE;
	  }
      generate_key_events (temp,0);
    }

}

void generate_key_events (struct keylist *kl, int MOD_FLAGS)
{
  BYTE i;
  DWORD UP;

  if (MOD_FLAGS&0x01)
    keybd_event(VK_SHIFT, shift_scan, 0,0);
  /* shift */
  if (MOD_FLAGS&0x02)
    keybd_event(VK_CONTROL, control_scan, 0,0);
  /* ctrl */
  if (MOD_FLAGS&0x04)
    keybd_event(VK_MENU, alt_scan, 0,0);
  /* alt */
  for (i=0;i<(int)kl->numevents;i++)
    {
      UP=0;
      if (kl->vkeys[i] &0x100)
	UP=KEYEVENTF_KEYUP;

      keybd_event ((BYTE)(kl->vkeys[i]&0xff),(BYTE)kl->skeys[i],UP,0);
    }
  if (MOD_FLAGS&0x01)
    keybd_event(VK_SHIFT, shift_scan, KEYEVENTF_KEYUP,0);
  /* shift */
  if (MOD_FLAGS&0x02)
    keybd_event(VK_CONTROL, control_scan,KEYEVENTF_KEYUP,0);
  /* ctrl */
  if (MOD_FLAGS&0x04)
    keybd_event(VK_MENU, alt_scan, KEYEVENTF_KEYUP,0);

#ifdef DELTEME
  vkmod = VkKeyScan (st[i]);
  vk = vkmod&0xff;
  scan  = MapVirtualKey(vk,0);

  keybd_event (vk,scan,0,0);
  keybd_event (vk,scan,KEYEVENTF_KEYUP,0);

#endif
}
void free_table()
{
  int i;
  for (i=0;i<8192;i++)
    {
      if (table[i])
	{
	  free(table[i]->vkeys);
	  free(table[i]->skeys);
	  free(table[i]);
	}
    }
  free(table);
}

void init_table()
{
  table = (struct keylist **) calloc (8192,sizeof(struct keylist *));
}

struct keylist *search_table  (int super_index)
{
  return table[super_index];
} 

HANDLE	 open_serial ()
     /* opens serial port to 2400,8N1 */
{
  BOOL BAD_SEQ,TWID_FOUND;
  HANDLE twid_port;
  char device [5];
  unsigned char lpBuffer[25];
  DCB dcb;
  int i,j,found=0,numread,first;
  COMMTIMEOUTS timeout;

  TWID_FOUND = FALSE;
  strcpy (device,"COM");
  device[4]=0;
  
 
  for (i=1;i<4; i++)
    {
      device [3] = (i+'0');
      twid_port=CreateFile(device, GENERIC_READ, 0,NULL,
			   OPEN_EXISTING, 0,NULL);
      if (twid_port == INVALID_HANDLE_VALUE)
	{
	  continue;
	}
      else
	{
	  if (!GetCommState (twid_port,&dcb))
	    {
	      CloseHandle (twid_port);
	      continue;
	    }
	  BuildCommDCB("baud=2400 parity=N data=8 stop=1",&dcb);
	  dcb.fRtsControl = 0;
	  dcb.fDtrControl=0;
	  /* get ready for reading */
	  /* set to 2400 baud */
	  if (!SetCommState (twid_port,&dcb))
	    {
	      CloseHandle (twid_port);
	      continue;
	    }
	  if (!GetCommTimeouts(twid_port,&timeout))
	    {
	      CloseHandle(twid_port);
	      continue;
	    }
	  timeout.ReadIntervalTimeout = 0;
	  timeout.ReadTotalTimeoutMultiplier = 20;
	  timeout.ReadTotalTimeoutConstant = 0;
	  if (!SetCommTimeouts (twid_port,&timeout))
	    {	    
	      CloseHandle (twid_port);
	      continue;
	    }
	  if (!ReadFile (twid_port, lpBuffer,20, &numread, NULL))
	    {
	      CloseHandle (twid_port);
	      continue;
	    }
	  else
	    {
	      /* analyze what we read and see if it looks like 
		 twiddle data */
	      if (numread!=20)
		{
		  CloseHandle (twid_port);
		  continue;
		}
	      for (j=0;(j<5 && !found); j++)
		{
		  /* find header byte */
		  if (!(lpBuffer[j]&0x80))
		    {
		      first =j;
		      found = TRUE;
		    }
		}
	      if (!found)
		{
			  
		  CloseHandle (twid_port);
		  continue;
		}
	      /* now look at the rest.. */
	      BAD_SEQ=FALSE;
	      for (j=first;(j<20 && !BAD_SEQ);j++)
		{
		  if ((j-first)%5 == 4)
		    {
		      /* check last byte*/
		      if ((lpBuffer[j]&0xf0) != 0x80)
			BAD_SEQ = TRUE;		     
		    }
		  else if ((j-first)%5)
		    {
		      /* 1-4 */
		      if (!(lpBuffer[j]&0x80))
			BAD_SEQ = TRUE;		    
		    }
		  else if (lpBuffer[j]&0x80)
		    BAD_SEQ=TRUE;
		}
	      if (BAD_SEQ)
		{
		  CloseHandle (twid_port);
		  continue;
		}
	    }	  
	}
      TWID_FOUND = TRUE;
      break;
      
      /* found the twiddler !*/
    }
  if (!TWID_FOUND)
    {
      twid_port = INVALID_HANDLE_VALUE;
      MessageBox (NULL, "Error!","Twiddler not detected",MB_ICONERROR|MB_OK);
    }

  return (twid_port);
}

int init(STATE states,int argc,char **argv)
{
  int i,initialize=0;
  char *filename;
  BYTE numscan;

  shift_scan = MapVirtualKey(VK_SHIFT,0);
  control_scan = MapVirtualKey(VK_CONTROL,0);
  alt_scan = MapVirtualKey(VK_MENU,0);
  init_table();
  for (i=0;i<keymaplength;i++)
    {
      keywordScan[i] = MapVirtualKey(keywordVirt[i],0);
    }

  /* let's see what the user has in store for us */
  for (i=1;i<argc;i++)
    {
      if (!strcmp (argv[i],"-f"))
	{
	  if (++i<argc)
	    {
	      filename = strdup (argv[i]);
	      if (read_file(filename)==-1)
		{
		  /*
		    printf ("Twiddler Keymap: Error reading default file %s !!\n",filename);
		    printf ("Check that file exists and try again\n");
		    printf ("Reverting to defaults..\n");*/
		}
	      else 
		{initialize =1;}
	    }
	}
      else if (!strcmp (argv[i],"-p"))
	{
	  /* specify port here */
	}
      else 
	{
	  /*printf ("Invalid argument...\n");
	    printf ("Usage: twid [-f filename] \n");
	    printf ("-f will specify a new default keymap\n");*/
	  /* return (0); skip it */    
	}
    }
 
  if (!initialize)
    if (read_file(DEFAULT_FILE)==-1)
      {
      	printf ("Twiddler Keymap: Error reading default file %s !!\n",DEFAULT_FILE);
	printf ("Check that file exists and try again\n");
	return(0);
      }

  /* inital states */
  if (GetKeyState (VK_NUMLOCK)==1)
    {
      /* turn that pesky num_lock off.. */
      numscan = (BYTE) MapVirtualKey (VK_NUMLOCK,0);
      keybd_event (VK_NUMLOCK,numscan,0,0);
      keybd_event (VK_NUMLOCK,numscan,KEYEVENTF_KEYUP,0);
    }

  states->cur_state.index=0;
  states->last_event.index=-1;
  states->last_state.index=0;
  mouse_filter = setup_filter (FILTER_LEVEL);
  twid_port = open_serial ();
  if (twid_port == INVALID_HANDLE_VALUE)
    {
      /******************************************
	real error stuff here 
	*********************/
      return (0);
    }
  return (1);
  /* OK, DONE! */
}

struct dxdy *setup_filter (int size)
{
  struct dxdy *temp;
  temp = calloc (size, sizeof (struct dxdy));
  filter_index =0;
  return temp;
}

int read_file (char *filename)
{
  int linenum,bucket=0;
  char line[MAXLINE];
  FILE *user;
 
  linenum=1;
  if (!(user = fopen (filename,"r")))
    return (-1);
  while (fgets(line,MAXLINE,user))
    {
      parse_line(line,linenum);
      linenum++;
    }
  fclose (user);
  return (1);
}

int parse_line (char *line,int linenum)
{
  char cur_string[MAXLINE],thumb[MAXLINE];
  UINT VirtKeys[MAXLINE],ScanCodes[MAXLINE],vkeytemp[MAXLINE],scantemp[MAXLINE];
  UINT savedv[MAXLINE],saveds[MAXLINE];
  int mapplace=0,index=0,i,t_index,savedstrokes=0,retval=1;
 
  /* format is :
     <thumb>+<thumb>... <four character code> = "string to translate"
     note the quotes, up to five entries (MOUSE is reserved) for thumb
     separated by a '+' */

  if (strlen(line)<2)
    return (-1);

  t_index=0;
  raise_case (line);
  sscanf(line,"%s",thumb);
  cropline (&line);
  
  do
    {
      i=get_thumb (thumb,cur_string,&t_index);
      if (!strcmp (cur_string, "CTRL"))
	index += CTRL_INDEX;
      else
	if (!strcmp(cur_string, "ALT"))
	  index += ALT_INDEX;
	else
	  if (!strcmp(cur_string, "NUM"))
	    index += NUM_INDEX;
	  else
	    if (!strcmp(cur_string, "FUNC"))
	      index += FUNC_INDEX;
	    else
	      if (!strcmp(cur_string, "SHIFT"))
		index += SHIFT_INDEX;
	      else
		if (!strcmp(cur_string, "0"));
		else
		  {
		    printf ("Error in config file, line %d: %s not a valid thumb button keyword\n",linenum,cur_string);
		    return (-1);
		  }
    } while (i);
  sscanf(line,"%s",cur_string);
  cropline (&line);
  if (strlen (cur_string)!=4)
    {
      printf ("Error in config file, line %d: Finger field %s is invalid\n",linenum,cur_string);
      return(-1);
    }
  else
    for (i=0;i<4; i++)
      {
	switch (cur_string[i])
	  {
	  case 'L':
	    index += LEFT_MASK << (2*i);
	    break;
	  case 'M': 
	    index += MID_MASK << (2*i);
	    break;
	  case 'R':
	    index += RIGHT_MASK << (2*i);
	    break;
	  case '0':
	    break;
	  default:
	    printf ("Error in config file,line %d: Finger field %s is invalid\n",linenum,cur_string);
	    return (-1);
	  }
      }
  
  sscanf(line,"%s",cur_string);
  cropline (&line);
  
  if ((strlen(cur_string)!=1) || (cur_string[0]!='='))
    {
      printf ("Error in config file, line %d: Equal sign Missing\n",linenum);
      return(-1);
    }

  /* three ways to get stuff: keywords, numbers, and strings */
  while ((line[0]==' ')||(line[0]=='\t'))
    line++;
  while (line[0]&&line[0]!='\n')
    {
      if (line[0]=='\"') 
	{ 
	  retval = 1;
	  if (!get_string (&line,scantemp,vkeytemp,linenum)) return (-1);
	  for (i=0;vkeytemp[i]!=0;i++)
	    {
	      VirtKeys[mapplace] = vkeytemp[i];
	      ScanCodes[mapplace] = scantemp[i];
	      mapplace++;
	    }
	  /* key downs */
	}
      else if ((line[0]>='0') && (line[0] <='9'))
	{ 
	  retval =1;
	  if (!get_num (&line,scantemp,vkeytemp,linenum)) return (-1); 
	  VirtKeys[mapplace] = vkeytemp[0];
	  ScanCodes[mapplace] = scantemp[0];
	  mapplace++;
	  VirtKeys[mapplace] = 0x100 | vkeytemp[0];
	  ScanCodes[mapplace] = scantemp[0];
	  mapplace++;
	}
      else 
	{ 

	  retval = get_keyword(&line,scantemp,vkeytemp,linenum);
	  if (!retval ) return (-1);
	  VirtKeys[mapplace] = vkeytemp[0];
	  ScanCodes[mapplace] = scantemp[0];
	  mapplace++;
	  if (retval==1)
	    {
	      VirtKeys[mapplace] = 0x100 | vkeytemp[0];
	      ScanCodes[mapplace] = scantemp[0];
	      mapplace++;
	    }
	  else
	    { 
	      savedv[savedstrokes] = vkeytemp[0];
	      saveds[savedstrokes] = scantemp[0];
	      savedstrokes++;
	    }

	}
      if (savedstrokes && retval==1)
	{
	  for (i=0;i<savedstrokes;i++)
	    {
	      /*key up the saved strokes */
	      VirtKeys[mapplace] = 0x100 | savedv[i];
	      ScanCodes[mapplace] = saveds[i];
	      mapplace++;
	    }
	  savedstrokes=0;
	}
      while ((line[0]==' ')||(line[0]=='\t'))
	line++;
    }
  table[index] = malloc (sizeof (struct keylist));
  table[index]->vkeys = calloc (mapplace,sizeof(UINT));
  table[index]->skeys = calloc (mapplace,sizeof(UINT));
  table[index]->numevents = mapplace;
  for  (i=0;i<mapplace;i++)
    {
      table[index]->vkeys[i] = VirtKeys[i];
      table[index]->skeys[i] = ScanCodes[i];
    }
  return (index);
}


void raise_case (char *line)
{
  int i,quote=0;
  /* puts the string to all CAPS except what's in quotes */
  for (i=0; i<(int)(strlen(line)); i++)
    {
      if (!quote)
	{
	  if (line[i]=='\"')
	    quote=1;
	  else if  ((line[i]>='a')&&(line[i]<='z'))
	    line [i] += ('A'-'a');
	}
      else
	if (line[i]=='\"')
	  quote=0;
    }
}

void cropline (char **line)
{
  /* gets rid of all whitespace, then one word (until next whitespace) */
  int flag=0;
  
  while (!(flag&&((**line==' ') || (**line=='\t'))))
    {
      if ((!flag) && ((**line != ' ') && (**line != '\t')))
	flag =1;
      (*line)++;
    }
}

int get_thumb (char *thumb,char *word,int *index)
{
  int begin=*index,i;

  while ((thumb[*index]!='\0') && (thumb[*index]!='+') &&
	 (thumb[*index]!='\t') && (thumb[*index]!=' '))
    (*index)++;
  
  for (i=0;i<10;i++)
    word[i]=0;
  for (i=begin;i<*index;i++)
    word[i-begin]=thumb[i];
  
  if (thumb[*index]=='+')
    {
      (*index)++;
      return (1);
    }
  else return (0);
}

int get_string(char** cur_string,UINT *scan,UINT *virtkey,int line)
{
  int index=0;
  int code=0;
  int skip=0;
  (*cur_string)++; 
  while (skip || (**cur_string!='\"'))
    {
      if ((!skip) && ((**cur_string)=='\\'))
	{
	  skip = 1;
	  (*cur_string)++;
	}
      else
	{
	  skip=0;
	  if (**cur_string=='\0') return (0);
	  code =VkKeyScan (**cur_string);
	  if (code&0x100)
	    {
	      virtkey[index] = VK_SHIFT;
	      scan[index++] = shift_scan;
	    }
	  if (code&0x200)
	    {
	      virtkey[index] = VK_CONTROL;
	      scan[index++] = control_scan;
	    }
	  if (code&0x400) 
	    {
	      virtkey[index] = VK_MENU;
	      scan[index++] = alt_scan;
	    }
	  virtkey[index] = code&0xff;
	  scan[index] = MapVirtualKey (virtkey[index],0);
	  index++;
	  virtkey[index] = 0x100 | (code&0xff);
	  scan[index] = scan[index-1];
	  index++;
	  if (code&0x100)
	    {
	      virtkey[index] = 0x100 | VK_SHIFT;
	      scan[index++] = shift_scan;
	    }
	  if (code&0x200)
	    {
	      virtkey[index] = 0x100 | VK_CONTROL;
	      scan[index++] = control_scan;
	    }
	  if (code&0x400) 
	    {
	      virtkey[index] = 0x100 | VK_MENU;
	      scan[index++] = alt_scan;
	    }
	  (*cur_string)++;
	}
    }
  virtkey[index]=0;
  (*cur_string)++;
  /* get rid of , */
  while ((**cur_string==' ')||(**cur_string=='\t'))
    (*cur_string)++;
  if (**cur_string==',')
    {
      (*cur_string)++;
      return (1);
    }
  if (**cur_string=='+')
    {
      (*cur_string)++;
      return (2);
    }
  if ((**cur_string=='\0')||(**cur_string=='\n'))
    return (1);
  printf ("Error in config file, line %d: String argument invalid\n",line);
  return(0);

}

int get_num (char** cur_string,UINT *scan,UINT *virtkey,int line)
{
  int num=0;
 
  while ((**cur_string>='0') && (**cur_string <='9'))
    {
      num *= 10;
      num += **cur_string-'0';
      (*cur_string)++;
    }
  *virtkey=num;
  *scan = MapVirtualKey(*virtkey,0);
  while ((**cur_string==' ')||(**cur_string=='\t'))
    (*cur_string)++;
  if (**cur_string==',')
    {
      (*cur_string)++;
      return (1);
    }
  if (**cur_string=='+')
    {
      (*cur_string)++;
      return (2);
    }
  if ((**cur_string=='\0')||(**cur_string=='\n'))
    return (1);
  printf ("Error in config file, line %d: Number argument invalid\n",line);
  return(0);
}

int get_keyword (char **cur_string, UINT *scan,UINT *virtkey,int line)
{
  int index,i,retval;
  char word[20]="";
  index=0;

  while (((*cur_string)[index]!='\0') && ((*cur_string)[index]!=',') 
	 && ((*cur_string)[index]!='+') &&
	 ((*cur_string)[index]!='\t') && ((*cur_string)[index]!=' ') &&
	 ((*cur_string)[index]!='\n'))
    index++;
  if ((*cur_string)[index] == '+')
    retval = 2;
  else
    retval=1;
  strncpy (word,*cur_string,index);
  
  if (((*cur_string)[index]=='\0')||((*cur_string)[index]=='\n'))
    *cur_string+=index;
  else *cur_string+=index+1;
  while ((**cur_string==' ')||(**cur_string=='\t'))
    (*cur_string)++; 
  for (i=0;i<keymaplength;i++)
    {
      if (!strcmp (word, keywordMap[i]))
	{
	  *scan = keywordScan[i];
	  *virtkey = keywordVirt [i];
	  return (retval);
	}
    }

  printf ("Error in config file, line %d: %s is an Invalid Keyword\n",line,word);
  return (0);
}

void ServiceStop()
{
  if ( hServerStopEvent )
    SetEvent(hServerStopEvent);
}
void quit ()
{
  CloseHandle (twid_port);
  free_table();
}

#ifdef REINIT_NOT_DONE_YET

int reinit (int signo)
{
  FILE *temp;
  char text [256];
  char *filename;

  if (!(filename = getenv ("TWID_DEFAULTS")))
    {
      filename = calloc (20,sizeof(char ));
      strcpy (filename , DEFAULT_FILE);
    }
  if (temp=fopen (TWID_TEMP,"r"))
    {
      fscanf (temp,"%s",text);
      fclose (temp);
      unlink (TWID_TEMP);
      free_table();
      init_table();
      if (read_file(filename)==-1)
	{
	  fprintf (stderr,"Twiddler Keymap: Error reading default file %s !!\n",filename);
	  fprintf (stderr,"Check that file exists and try again\n");
	  return(0);
	}
      fprintf (stdout,"\nTwiddler Defaults restored.\n");
      fflush (stdout);
      if (strcmp(text,filename))
	{
	  if (read_file(text)==-1)
	    {
	      fprintf (stderr,"Twiddler Keymap: Error reading file %s!\n",text);
	      fprintf (stderr,"Additions will not be active...\n");
	    }
	}
    }
  else
    {
      fprintf (stderr,"Could not open message file\n");
    }
  return (1);
}

#endif
