/* Twiddler Driver for Windows NT (Test User Mode Version)
 * 1/15/97
 * written by:Jeff Levine
 * <tornado@media.mit.edu>
 * for Handykey Corporation 
*/
/* Twiddler.h file */

#include <stdio.h>
#include <stdlib.h>
#include <sys/types.h>
#include <math.h>
#include <string.h>
#include <windows.h>
#include "service.h"

struct data
{
  int index;
  int bucket;
  int horiz_tilt, vert_tilt;  
};

struct keylist
{
	BYTE numevents;
	UINT *skeys;
	UINT *vkeys;
};

struct dxdy
{
  int dxraw, dyraw;
  int dx;
  int dy;
  struct dxdy *next;
  struct dxdy *prev;
};

typedef struct data DATA;

struct ent
{ 
  int index;
  char* str;
};
typedef struct ent entry;

struct state
{
  DATA last_state,cur_state,last_event;
  int repeat_count; 
  int time_count;
  char flags;
  /* FLAGS FORMAT:
  (LSB) 0  events_enabled 
        1  repeat_enabled
        2  NUM_LOCK
	    3  
        4  
        5  
        6  ?
        7  ?
	*/ 
};
typedef struct state *STATE;

/* Function Def's */
void get3mean (struct dxdy *filter, int *cx, int *cy, int newcx, int newcy,
	       int *index, int filter_size);
void add_to_filter (struct dxdy *filter, int dx, int dy, int cx,int cy,
		    int *index, int filter_size);

void mainloop();
void cropline (char **line);
void processbuffer (unsigned char *buffer);
void getport (char *arg,char *port);
HANDLE open_serial ();
void handle_mouse_activity (int index, unsigned char *buffer,int fil_lev);
void generate_key_events (struct keylist *kl, int MOD_FLAGS);
int get_thumb (char *cur_string,char *word,int *index);
int get_keyword (char **cur_string,UINT *scan,UINT *virt,int line);
int get_num (char **cur_string,UINT *scan, UINT *virt,int line);
int get_string(char **cur_string,UINT *scan,UINT *virt,int line);
int parse_line (char *line, int linenum);
int init(STATE states,int argc,char **argv);
int update (void);
int read_file (char *filename);
void get_text ();
void raise_case (char *line);
void free_table (void);
void init_table (void);
void quit ();
struct keylist *search_table(int super_index);
void insert (int bucket,entry *map);
void getXresolution (int *x_res, int *y_res);
void kill_filter (struct dxdy *filter, int filter_level);
struct dxdy *setup_filter (int size);
void filter_mouse (struct dxdy *filter, int filter_level,
			   int res_level, int *dx, int *dy, int *index,
			   int filter_size);
int reinit (int signo);
/* DEFINES */


#define DEFAULT_FILE "c:\\twiddler\\twid_defaults.ini"
#define MAXLINE 120 

#define SHIFT_INDEX 0x000100
#define NUM_INDEX   0x000200
#define FUNC_INDEX  0x000400
#define CTRL_INDEX  0x000800
#define ALT_INDEX   0x001000
#define MOUSE_INDEX 0x002000
#define LEFT_MASK   1
#define MID_MASK    2
#define RIGHT_MASK  3

#define M_OFF 0
#define M_FIRST_HIT 1
#define M_ON 2
#define M_2ND_HIT 3

#define LEFT_BUTTON 3
#define MID_BUTTON 2
#define RIGHT_BUTTON 1

#define MOUSE_REVERSE 0x54
#define MOUSE_HIRES 0x30
#define MOUSE_LORES 0x20
#define M_DOUBLE_CLICK 12

#define EVENT_ENABLE 0x01
#define REPEAT_ENABLE 0x02
#define NUM_LOCK 0x04


struct norep
{
  int num;
  UINT *code;
};
struct norep norepeat;


#define REPEAT_THRESH 15 /* at 2400 baud this is about 1/3 of a second */
#define REPEAT_TIME_THRESH 20

#define SCROLL_ON_INDEX 2113
#define SCROLL_OFF_INDEX 2098
#define BACKSPACE_CODE 8
#define ENTER_CODE 13
#define ESCAPE_CODE 27
#define TAB_CODE 9
#define FILTER_LEVEL 15

#define keymaplength 61

 char *keywordMap [] = {
"BACKSPACE","ENTER","TAB","CONTROL","ALT","DELETE","SHIFT","ESCAPE",
"NUM_LOCK","CAPS_LOCK","SCROLL_LOCK","ARROW_UP","ARROW_DOWN","ARROW_LEFT","ARROW_RIGHT",
"F1","F2","F3","F4","F5","F6","F7","F8","F9","F10","F11","F12","F13","F14","F15","F16","F17",
"F18","F19","F20","F21","F22","F23","F24","PRINT_SCR","PAUSE","BREAK","INSERT","HOME","PG_UP","PG_DOWN",
"END","NUM_1","NUM_2","NUM_3","NUM_4","NUM_5","NUM_6","NUM_7","NUM_8","NUM_9","NUM_0",
"NUM_TIMES","NUM_DIVIDE","NUM_MINUS","NUM_PLUS","NUM_ENTER","NUM_PERIOD"};

UINT keywordVirt [] = {
VK_BACK,VK_RETURN,VK_TAB,VK_CONTROL,VK_MENU,VK_DELETE,VK_SHIFT,VK_ESCAPE,
VK_NUMLOCK,VK_CAPITAL,VK_SCROLL,VK_UP,VK_DOWN,VK_LEFT,VK_RIGHT,VK_F1,VK_F2,VK_F3,VK_F4,VK_F5,VK_F6,VK_F7,
VK_F8,VK_F9,VK_F10,VK_F11,VK_F12,VK_F13,VK_F14,VK_F15,VK_F16,VK_F17,VK_F18,VK_F19,VK_F20,VK_F21,
VK_F22,VK_F23,VK_F24,VK_SNAPSHOT,VK_PAUSE,VK_CANCEL,VK_INSERT,VK_HOME,VK_PRIOR,VK_NEXT,VK_END,
VK_NUMPAD1,VK_NUMPAD2,VK_NUMPAD3,VK_NUMPAD4,VK_NUMPAD5,VK_NUMPAD6,VK_NUMPAD7,VK_NUMPAD8,VK_NUMPAD9,
VK_NUMPAD0,VK_MULTIPLY,VK_DIVIDE,VK_SUBTRACT,VK_ADD,VK_SEPARATOR,VK_DECIMAL};

UINT keywordScan [ keymaplength];






